//345678901234567890123456789012345678901234567890123456789
//Note: this is wide format for small fonts.
//=======================================================//

//Be sure to preserve < and > in html version
//01/19/06 ready for publication.

//Apparently this file has a problem. See the file named ImgMod35Good.java

/*File ImgMod35.java
Copyright 2005, R.G.Baldwin

This program performs a forward DCT on an image converting
the three color planes into spectral planes.  Then it
performs an inverse DCT on the three spectral planes
converting them back into image color planes.

Nothing is done to the spectral planes following the
forward DCT and before the inverse DCT.  However,
additional processing, such as requantization and
compression, followed by decompression could be inserted at
that point.

This is an update to ImgMod34.  This update breaks the
image down into blocks of 8x8-pixels and performs the
forward DCT on each block producing 8x8 blocks of spectral
coefficient data.  Then it performs an inverse DCT on each
of the 8x8 spectral blocks, reproducing the 8x8 pixel
blocks.  The composite of all the 8x8-pixel blocks
constitutes the total image.  Zero padding is applied to
the right and bottom of the image to force each
dimension of the image to be a multiple of 8 pixels.  This
padding is trimmed from the resulting spectral planes and
the resulting composite image before the processed
composite image is returned to the calling method.  This
process doesn't appear to have a detrimental impact on the
quality of the image at the edges.

For illustration purposes only, the method named
inverseXform8x8Plane contains a statement that can be
activated to cause the 8x8 block structure of the entire
process to become visible in the final image. This feature
should be disabled by default because it badly corrupts the
visual quality of the image when it is enabled.

Note that this version of the program is significantly
faster than the version named ImgMod034, which does not
break the image down into 8x8-pixel blocks, but rather does
the forward and inverse DCT on the entire image.

Among other things, this program uses an 8x8 cosine lookup
table instead of computing the cosine value every time it
is needed.  This is probably one factor in the speed
improvement.  Another factor is the simple fact that less
arithmetic is required to perform the transform when the
image is sub-divided into blocks and the transforms are
performed on the individual blocks instead of transforming
the entire image as a whole.

The class is designed to be driven by the class named
ImgMod02a.

Enter the following at the command line to run this
program:

java ImgMod02a ImgMod35 ImageFileName

where ImageFileName is the name of a .gif or .jpg file,
including the extension.

When you click the Replot button, the process will be
repeated and the results will be re-displayed.  Because
there is no opportunity for user input after the program is
started, the Replot button is of little value to this
program.

This program requires access to the following class files
plus some inner classes that are defined inside the
following classes:

ImgIntfc02.class
ImgMod02a.class
ImgMod35.class
InverseDCT02.class
ForwardDCT02.class

Tested using J2SE 5.0 and WinXP.  J2SE 5.0 or later is
required due to the use of static imports.
**********************************************************/
import java.awt.*;
import java.io.*;
import static java.lang.Math.*;

class ImgMod35 implements ImgIntfc02{

  //This method is required by ImgIntfc02.  It is called at
  // the beginning of the run and each time thereafter that
  // the user clicks the Replot button on the Frame
  // contaning the images.
  public int[][][] processImg(int[][][] threeDPix,
                              int imgRows,
                              int imgCols){

    //Create an empty output array of the same size as the
    // incoming array.
    int[][][] output = new int[imgRows][imgCols][4];

    //Make a working copy of the 3D pixel array as type
    // double to avoid making permanent changes to the
    // original image data.  Also, all processing will be
    // performed as type double.
    double[][][] working3D = copyToDouble(threeDPix);

    //The following code can be enabled to set any of the
    // three colors to black, thus removing them from the
    // output.
    for(int row = 0;row < imgRows;row++){
      for(int col = 0;col < imgCols;col++){
//        working3D[row][col][1] = 0;
//        working3D[row][col][2] = 0;
//        working3D[row][col][3] = 0;
      }//end inner loop
    }//end outer loop

    //Extract and do a forward DCT on the red color plane
    double[][] redPlane = extractPlane(working3D,1);
    //Expand the plane such that both dimensions are a
    // multiple of 8.
    redPlane = expandPlane(redPlane);
    //Do the forward DCT on the redPlane turning it into a
    // spectral plane where each individual spectrum stored
    // in the plane is an 8x8 square.  To see the
    // individual spectra, disable all but one color plane
    // and disable the inverse transforms that are
    // performed later.
    forwardXformPlane(redPlane);
    //Insert the spectral plane back into the 3D array.
    // This method also trims off any extra rows and
    // columns created by expanding the dimensions to a
    // multiple of 8.  Since those rows and columns
    // contain spectral data, the trimming could
    // conceivably cause problems with the quality of the
    // image at the right and bottom edges later.  If so, I
    // will need to save the spectral plane somewhere else
    // without trimming it.
    insertPlane(working3D,redPlane,1);

    //Extract and do a forward DCT on the green color
    // plane.
    double[][] greenPlane = extractPlane(working3D,2);
    greenPlane = expandPlane(greenPlane);
    forwardXformPlane(greenPlane);
    //Insert the plane back into the 3D array
    insertPlane(working3D,greenPlane,2);

    //Extract and do a forward DCT on the blue color plane.
    double[][] bluePlane = extractPlane(working3D,3);
    bluePlane = expandPlane(bluePlane);
    forwardXformPlane(bluePlane);
    //Insert the plane back into the 3D array
    insertPlane(working3D,bluePlane,3);

    //All three color planes have now been transformed into
    // spectral planes where each spectral plane is
    // composed of a large number of contiguous 8x8
    // spectral data blocks.

    //Now transform the spectral planes back into image
    // color planes.
    //You can disable this section of code and disable two
    // of the three colors to get some idea as to what the
    // spectral planes look like.  If you don't disable two
    // of the colors, you will see a composite of all three
    // spectral planes, which may or may not be
    // informative.  At least, it shows the 8x8 block
    // structure of the spectral planes even if it is a
    // composite of three spectral planes.

    //Extract and do an inverse DCT on the red spectral
    // plane to produce a color image plane.
    redPlane = extractPlane(working3D,1);
    //Expand the plane such that both dimensions are a
    // multiple of 8.
    redPlane = expandPlane(redPlane);
    //Do the inverse DCT on the spectral plane producing a
    // color image plane.
    inverseXformPlane(redPlane);
    //Insert the color plane back into the 3D array. This
    // also trims off any extra rows and columns created by
    // expanding to a multiple of 8.
    insertPlane(working3D,redPlane,1);

    //Extract and do an inverse DCT on the green spectral
    // plane to produce a color image plane.
    greenPlane = extractPlane(working3D,2);
    greenPlane = expandPlane(greenPlane);
    inverseXformPlane(greenPlane);
    insertPlane(working3D,greenPlane,2);

    //Extract and do an inverse DCT on the blue spectral
    // plane to produce a color image plane.
    bluePlane = extractPlane(working3D,3);
    bluePlane = expandPlane(bluePlane);
    inverseXformPlane(bluePlane);
    insertPlane(working3D,bluePlane,3);

    //End of section that does the inverse transforms.

    //Convert the image color planes to type int and return
    // the array of pixel data to the calling method.
    output = copyToInt(working3D);
    //Return a reference to the output array.
    return output;

  }//end processImg method
  //-----------------------------------------------------//

  //The purpose of this method is to extract a specified
  // row from a double 2D plane and to return it as a one-
  // dimensional array of type double.
  double[] extractRow(double[][] colorPlane,int row){

    int numCols = colorPlane[0].length;
    double[] output = new double[numCols];
    for(int col = 0;col < numCols;col++){
      output[col] = colorPlane[row][col];
    }//end outer loop
    return output;
  }//end extractRow
  //-----------------------------------------------------//

  //The purpose of this method is to insert a specified
  // row of double data into a double 2D plane.
  void insertRow(double[][] colorPlane,
                 double[] theRow,
                 int row){
    int numCols = colorPlane[0].length;
    double[] output = new double[numCols];
    for(int col = 0;col < numCols;col++){
      colorPlane[row][col] = theRow[col];
    }//end outer loop
  }//end insertRow
  //-----------------------------------------------------//

  //The purpose of this method is to extract a specified
  // col from a double 2D plane and to return it as a one-
  // dimensional array of type double.
  double[] extractCol(double[][] colorPlane,int col){
    int numRows = colorPlane.length;
    double[] output = new double[numRows];
    for(int row = 0;row < numRows;row++){
      output[row] = colorPlane[row][col];
    }//end outer loop
    return output;
  }//end extractCol
  //-----------------------------------------------------//

  //The purpose of this method is to insert a specified
  // col of double data into a double 2D color plane.
  void insertCol(double[][] colorPlane,
                 double[] theCol,
                 int col){
    int numRows = colorPlane.length;
    double[] output = new double[numRows];
    for(int row = 0;row < numRows;row++){
      colorPlane[row][col] = theCol[row];
    }//end outer loop
  }//end insertCol
  //-----------------------------------------------------//

  //The purpose of this method is to extract a color plane
  // from the double version of an image and to return it
  // as a 2D array of type double.
  public double[][] extractPlane(
                              double[][][] threeDPixDouble,
                              int plane){

    int numImgRows = threeDPixDouble.length;
    int numImgCols = threeDPixDouble[0].length;

    //Create an empty output array of the same
    // size as a single plane in the incoming array of
    // pixels.
    double[][] output =new double[numImgRows][numImgCols];

    //Copy the values from the specified plane to the
    // double array.
    for(int row = 0;row < numImgRows;row++){
      for(int col = 0;col < numImgCols;col++){
        output[row][col] =
                          threeDPixDouble[row][col][plane];
      }//end loop on col
    }//end loop on row
    return output;
  }//end extractPlane
  //-----------------------------------------------------//

  //The purpose of this method is to insert a double 2D
  // plane into the double 3D array that represents an
  // image.  This method also trims off any extra rows and
  // columns in the double 2D plane.
  public void insertPlane(double[][][] threeDPixDouble,
                          double[][] colorPlane,
                          int plane){

    int numImgRows = threeDPixDouble.length;
    int numImgCols = threeDPixDouble[0].length;

    //Copy the values from the incoming color plane to the
    // specified plane in the 3D array.
    for(int row = 0;row < numImgRows;row++){
      for(int col = 0;col < numImgCols;col++){
        threeDPixDouble[row][col][plane] =
                                      colorPlane[row][col];
      }//end loop on col
    }//end loop on row
  }//end insertPlane
  //-----------------------------------------------------//

  //This method copies an int version of a 3D pixel array
  // to an new pixel array of type double.
  double[][][] copyToDouble(int[][][] threeDPix){
    int imgRows = threeDPix.length;
    int imgCols = threeDPix[0].length;

    double[][][] new3D = new double[imgRows][imgCols][4];
    for(int row = 0;row < imgRows;row++){
      for(int col = 0;col < imgCols;col++){
        new3D[row][col][0] = threeDPix[row][col][0];
        new3D[row][col][1] = threeDPix[row][col][1];
        new3D[row][col][2] = threeDPix[row][col][2];
        new3D[row][col][3] = threeDPix[row][col][3];
      }//end inner loop
    }//end outer loop
    return new3D;
  }//end copyToDouble
  //-----------------------------------------------------//

  //This method copies a double version of a 3D pixel array
  // into a new pixel array of type int.
  int[][][] copyToInt(double[][][] threeDPixDouble){
    int imgRows = threeDPixDouble.length;
    int imgCols = threeDPixDouble[0].length;

    int[][][] new3D = new int[imgRows][imgCols][4];
    for(int row = 0;row < imgRows;row++){
      for(int col = 0;col < imgCols;col++){
        new3D[row][col][0] =
                         (int)threeDPixDouble[row][col][0];
        new3D[row][col][1] =
                         (int)threeDPixDouble[row][col][1];
        new3D[row][col][2] =
                         (int)threeDPixDouble[row][col][2];
        new3D[row][col][3] =
                         (int)threeDPixDouble[row][col][3];
      }//end inner loop
    }//end outer loop
    return new3D;
  }//end copyToInt
  //-----------------------------------------------------//

  //The purpose of this method is to clip all negative
  // color values in a double color plane to a value of 0.
  void clipToZero(double[][] colorPlane){
    int numImgRows = colorPlane.length;
    int numImgCols = colorPlane[0].length;
    //Do the clip
    for(int row = 0;row < numImgRows;row++){
      for(int col = 0;col < numImgCols;col++){
        if(colorPlane[row][col] < 0){
          colorPlane[row][col] = 0;
        }//end if
      }//end inner loop
    }//end outer loop
  }//end clipToZero
  //-----------------------------------------------------//

  //The purpose of this method is to clip all color values
  // in a double color plane that are greater than 255 to
  // a value of 255.
  void clipTo255(double[][] colorPlane){
    int numImgRows = colorPlane.length;
    int numImgCols = colorPlane[0].length;
    //Do the clip
    for(int row = 0;row < numImgRows;row++){
      for(int col = 0;col < numImgCols;col++){
        if(colorPlane[row][col] > 255){
          colorPlane[row][col] = 255;
        }//end if
      }//end inner loop
    }//end outer loop
  }//end clipTo255
  //-----------------------------------------------------//

  //Expand a double 2D plane such that both dimensions are
  // a multiple of 8.
  double[][] expandPlane(double[][] plane){
    int rows = plane.length;
    int cols = plane[0].length;
    double[][] output;

    int rowsRem = rows%8;
    int colsRem = cols%8;

    if((rowsRem == 0) && (colsRem == 0)){
      //No expansion is required.
      return plane;
    }else{
      //An expansion is required.  Copy the data from the
      // incoming array to a new expanded array, leaving
      // pad values of 0.0 at the right and bottom edges.
      output = new double
              [rows + (8 - rowsRem)][cols + (8 - colsRem)];
      for(int row = 0;row < rows;row++){
        for(int col = 0;col < cols;col++){
          output[row][col] = plane[row][col];
        }//end inner loop
      }//end outer loop
      return output;
    }//end else
  }//end expandPlane
  //-----------------------------------------------------//

  //Extracts and returns an 8x8 block from a double 2D
  // plane.
  double[][] get8x8Block(double[][] colorPlane,
                         int segRow,
                         int segCol){
    double[][] the8x8Block = new double[8][8];
    for(int bigRow = segRow * 8,smallRow = 0;
                                 bigRow < (segRow * 8 + 8);
                                      bigRow++,smallRow++){
      for(int bigCol = segCol * 8,smallCol = 0;
                                   bigCol < segCol * 8 + 8;
                                      bigCol++,smallCol++){
        the8x8Block[smallRow][smallCol] =
                                colorPlane[bigRow][bigCol];
      }//end inner loop
    }//end outer loop
    return the8x8Block;
  }//end get8x8Block
  //-----------------------------------------------------//

  //Inserts an 8x8 block into a double 2D plane
  void insert8x8Block(double[][] colorPlane,
                      double[][] the8x8Plane,
                      int segRow,
                      int segCol){
    for(int bigRow = segRow * 8,smallRow = 0;
                                 bigRow < (segRow * 8 + 8);
                                      bigRow++,smallRow++){
      for(int bigCol = segCol * 8,smallCol = 0;
                                   bigCol < segCol * 8 + 8;
                                      bigCol++,smallCol++){
        colorPlane[bigRow][bigCol] =
                           the8x8Plane[smallRow][smallCol];
      }//end inner loop
    }//end outer loop
  }//end insert8x8Block
  //-----------------------------------------------------//

  //Breaks a double color plane down into 8x8-pixel blocks
  // and does a forward DCT xform on each block.  Assembles
  // the resulting spectral data blocks into a spectral
  // plane. Assumes that the dimensions of the color plane
  // are multiples of 8
  void forwardXformPlane(double[][] plane){
    int pixRows = plane.length;
    int pixCols = plane[0].length;
    //Loop on rows of 8x8 blocks
    for(int segRow = 0;segRow < pixRows/8;segRow++){
      //Loop on cols of 8x8 blocks
      for(int segCol = 0;segCol < pixCols/8;segCol++){
        double[][] the8x8Plane =
                          get8x8Block(plane,segRow,segCol);
        forwardXform8x8Block(the8x8Plane);
        insert8x8Block(plane,the8x8Plane,segRow,segCol);
      }//end inner loop
    }//end outer loop
  }//end forwardXformPlane
  //-----------------------------------------------------//

  //This method does a forward DCT on an 8x8 block of
  // pixels from a color plane received as an incoming
  // parameter.
  void forwardXform8x8Block(double[][] the8x8Block){

    int imgRows = 8;
    int imgCols = 8;

    //Extract each row from the 8x8-pixel block and perform
    // a forward DCT on the row. Then insert the
    // transformed row back into the block.  At that point,
    // the row no longer contains color pixel data, but
    // has been transformed into a row of spectral data.
    for(int row = 0;row < imgRows;row++){
      double[] theRow = extractRow(the8x8Block,row);

      double[] theXform = new double[theRow.length];
      //Perform the forward transform.
      //ForwardDCT02.transform(theRow,theXform);

      //Insert the transformed row back into the block.
      insertRow(the8x8Block,theXform,row);
    }//end for loop

    //The block now contains the results of doing the
    // horizontal DCT one row at a time.  The block no
    // longer contains color pixel data.  Rather, it
    // contains spectral data for the horizontal dimension
    // only.

    //Extract each column from the block and perform a
    // forward DCT on the column. Then insert the
    // transformed column back into the block.
    for(int col = 0;col < imgCols;col++){
      double[] theCol = extractCol(the8x8Block,col);

      double[] theXform = new double[theCol.length];
      ForwardDCT02.transform(theCol,theXform);

      insertCol(the8x8Block,theXform,col);
    }//end for loop

    //The 8x8 block has now been converted into an 8x8
    // block of spectral coefficient data.

  }//end forwardXform8x8Block
  //-----------------------------------------------------//

  //Breaks a spectral plane down into 8x8 blocks, performs
  // an inverse DCT on each block, and inserts the
  // resulting 8x8 image blocks into a color plane. Assumes
  // that the dimensions of the plane are multiples of 8
  void inverseXformPlane(double[][] plane){
    int pixRows = plane.length;
    int pixCols = plane[0].length;
    //Loop on rows of 8x8 blocks
    for(int segRow = 0;segRow < pixRows/8;segRow++){
      //Loop on cols of 8x8 blocks
      for(int segCol = 0;segCol < pixCols/8;segCol++){
        double[][] the8x8Block =
                          get8x8Block(plane,segRow,segCol);
        inverseXform8x8Plane(the8x8Block);
        insert8x8Block(plane,the8x8Block,segRow,segCol);
      }//end inner loop
    }//end outer loop
  }//end inverseXformPlane
  //-----------------------------------------------------//

  //This method performs an inverse DCT on an 8x8 spectral
  // data block.
  void inverseXform8x8Plane(double[][] the8x8Block){

    int imgRows = 8;
    int imgCols = 8;

    //Extract each col from the spectral data block and
    // perform an inverse DCT on the column. Then insert it
    // back into the block, which is being transformed into
    // a block of color pixel data.
    for(int col = 0;col < imgCols;col++){
      double[] theXform = extractCol(the8x8Block,col);

      double[] theCol = new double[theXform.length];
      //Now perform the inverse transform.
      InverseDCT02.transform(theXform,theCol);

      //Insert it back into the block.
      insertCol(the8x8Block,theCol,col);
    }//end for loop

    //At this point, an inverse DCT has been performed on
    // each column in the spectral data block, one column
    // at a time.

    //Extract each row from the block and perform an
    // inverse DCT on the row. Then insert it back into the
    // block.  The row now contains color pixel data.
    for(int row = 0;row < imgRows;row++){
      double[] theXform = extractRow(the8x8Block,row);

      double[] theRow = new double[theXform.length];
      //Now perform the inverse transform.
      InverseDCT02.transform(theXform,theRow);

      //Insert it back into the block.
      insertRow(the8x8Block,theRow,row);
    }//end for loop

    //For illustration purposes only, activate the
    // following statement to cause the 8x8 block
    // structure of the entire process to become visible
    // in the displayed image.  This causes each individual
    // block of pixel data to be a slightly different
    // color.
    //addRandom(the8x8Block);

    //At this point, the spectral data block has been
    // converted into an 8x8 block of pixel color data.
    // Ultimately it will be necessary to convert it to
    // 8-bit unsigned pixel color format in order to
    // display it.  Clip to zero and 255.
    clipToZero(the8x8Block);
    clipTo255(the8x8Block);
  }//end inverseXform8x8Plane
  //-----------------------------------------------------//

  //The purpose of this method is to add the same random
  // color value to all of the values in an 8x8 block of
  // image color data to cause the block structure of the
  // process to be visible.  This capability is intended
  // to be used for illustration purposes only as it
  // badly corrupts the quality of the image.
  void addRandom(double[][] colorPlane){
    double bias = 100.0 * random();
    int numImgRows = colorPlane.length;
    int numImgCols = colorPlane[0].length;
    //Do the clip
    for(int row = 0;row < numImgRows;row++){
      for(int col = 0;col < numImgCols;col++){
          colorPlane[row][col] += bias;
      }//end inner loop
    }//end outer loop
  }//end addRandom
  //-----------------------------------------------------//
}//end class ImgMod35